@@ -1,6 +1,6 @@
 /* FAudio - XAudio Reimplementation for FNA
  *
- * Copyright (c) 2011-2021 Ethan Lee, Luigi Auriemma, and the MonoGame Team
+ * Copyright (c) 2011-2023 Ethan Lee, Luigi Auriemma, and the MonoGame Team
  *
  * This software is provided 'as-is', without any express or implied warranty.
  * In no event will the authors be held liable for any damages arising from
@@ -89,6 +89,10 @@ uint32_t FACTAudioEngine_Release(FACTAudioEngine *pEngine)
 	FAudio_PlatformDestroyMutex(pEngine->wbLock);
 	FAudio_PlatformUnlockMutex(pEngine->apiLock);
 	FAudio_PlatformDestroyMutex(pEngine->apiLock);
+	if (pEngine->settings != NULL)
+	{
+		pEngine->pFree(pEngine->settings);
+	}
 	pEngine->pFree(pEngine);
 	FAudio_PlatformRelease();
 	return 0;
@@ -163,13 +167,81 @@ uint32_t FACTAudioEngine_Initialize(
 
 	FAudio_PlatformLockMutex(pEngine->apiLock);
 
-	/* Parse the file */
-	parseRet = FACT_INTERNAL_ParseAudioEngine(pEngine, pParams);
-	if (parseRet != 0)
+	if (!pParams->pGlobalSettingsBuffer || pParams->globalSettingsBufferSize == 0)
 	{
-		FAudio_PlatformUnlockMutex(pEngine->apiLock);
-		return parseRet;
+		/* No file? Just go with a safe default. (Also why are you using XACT) */
+		pEngine->categoryCount = 3;
+		pEngine->variableCount = 0;
+		pEngine->rpcCount = 0;
+		pEngine->dspPresetCount = 0;
+		pEngine->dspParameterCount = 0;
+
+		pEngine->categories = (FACTAudioCategory*) pEngine->pMalloc(
+			sizeof(FACTAudioCategory) * pEngine->categoryCount
+		);
+		pEngine->categoryNames = (char**) pEngine->pMalloc(
+			sizeof(char*) * pEngine->categoryCount
+		);
+
+		pEngine->categoryNames[0] = pEngine->pMalloc(7);
+		FAudio_strlcpy(pEngine->categoryNames[0], "Global", 7);
+		pEngine->categories[0].instanceLimit = 255;
+		pEngine->categories[0].fadeInMS = 0;
+		pEngine->categories[0].fadeOutMS = 0;
+		pEngine->categories[0].maxInstanceBehavior = 0;
+		pEngine->categories[0].parentCategory = -1;
+		pEngine->categories[0].volume = 1.0f;
+		pEngine->categories[0].visibility = 1;
+		pEngine->categories[0].instanceCount = 0;
+		pEngine->categories[0].currentVolume = 1.0f;
+
+		pEngine->categoryNames[1] = pEngine->pMalloc(8);
+		FAudio_strlcpy(pEngine->categoryNames[1], "Default", 8);
+		pEngine->categories[1].instanceLimit = 255;
+		pEngine->categories[1].fadeInMS = 0;
+		pEngine->categories[1].fadeOutMS = 0;
+		pEngine->categories[1].maxInstanceBehavior = 0;
+		pEngine->categories[1].parentCategory = 0;
+		pEngine->categories[1].volume = 1.0f;
+		pEngine->categories[1].visibility = 1;
+		pEngine->categories[1].instanceCount = 0;
+		pEngine->categories[1].currentVolume = 1.0f;
+
+		pEngine->categoryNames[2] = pEngine->pMalloc(6);
+		FAudio_strlcpy(pEngine->categoryNames[2], "Music", 6);
+		pEngine->categories[2].instanceLimit = 255;
+		pEngine->categories[2].fadeInMS = 0;
+		pEngine->categories[2].fadeOutMS = 0;
+		pEngine->categories[2].maxInstanceBehavior = 0;
+		pEngine->categories[2].parentCategory = 0;
+		pEngine->categories[2].volume = 1.0f;
+		pEngine->categories[2].visibility = 1;
+		pEngine->categories[2].instanceCount = 0;
+		pEngine->categories[2].currentVolume = 1.0f;
+
+		pEngine->variables = NULL;
+		pEngine->variableNames = NULL;
+		pEngine->globalVariableValues = NULL;
+		pEngine->rpcs = NULL;
+		pEngine->dspPresets = NULL;
 	}
+	else
+	{
+		/* Parse the file */
+		parseRet = FACT_INTERNAL_ParseAudioEngine(pEngine, pParams);
+		if (parseRet != 0)
+		{
+			FAudio_PlatformUnlockMutex(pEngine->apiLock);
+			return parseRet;
+		}
+	}
+
+	/* Peristent Notifications */
+	pEngine->notifications = 0;
+	pEngine->cue_context = NULL;
+	pEngine->sb_context = NULL;
+	pEngine->wb_context = NULL;
+	pEngine->wave_context = NULL;
 
 	/* Assign the callbacks */
 	pEngine->notificationCallback = pParams->fnNotificationCallback;
@@ -218,6 +290,7 @@ uint32_t FACTAudioEngine_Initialize(
 			NULL
 		) != 0) {
 			FAudio_Release(pEngine->audio);
+			FAudio_PlatformUnlockMutex(pEngine->apiLock);
 			return FAUDIO_E_INVALID_CALL;
 		}
 	}
@@ -280,6 +353,16 @@ uint32_t FACTAudioEngine_ShutDown(FACTAudioEngine *pEngine)
 		FAudio_StopEngine(pEngine->audio);
 	}
 
+	/* Purge All pending notifactions */
+	while (pEngine->wb_notifications_list)
+	{
+		FACTNotification *note = (FACTNotification*) pEngine->wb_notifications_list->entry;
+		pEngine->notificationCallback(note);
+		LinkedList_RemoveEntry(&pEngine->wb_notifications_list, note, pEngine->apiLock, pEngine->pFree);
+	}
+
+	pEngine->notifications = 0;
+
 	/* This method destroys all existing cues, sound banks, and wave banks.
 	 * It blocks until all cues are destroyed.
 	 */
@@ -361,9 +444,17 @@ uint32_t FACTAudioEngine_DoWork(FACTAudioEngine *pEngine)
 	uint8_t i;
 	FACTCue *cue;
 	LinkedList *list;
+	FACTNotification *note;
 
 	FAudio_PlatformLockMutex(pEngine->apiLock);
 
+	while (pEngine->wb_notifications_list)
+	{
+		note = (FACTNotification*) pEngine->wb_notifications_list->entry;
+		pEngine->notificationCallback(note);
+		LinkedList_RemoveEntry(&pEngine->wb_notifications_list, note, pEngine->apiLock, pEngine->pFree);
+	}
+
 	list = pEngine->sbList;
 	while (list != NULL)
 	{
@@ -423,6 +514,7 @@ uint32_t FACTAudioEngine_CreateInMemoryWaveBank(
 	uint32_t dwAllocAttributes,
 	FACTWaveBank **ppWaveBank
 ) {
+	FACTNotification *note;
 	uint32_t retval;
 	FAudio_PlatformLockMutex(pEngine->apiLock);
 	retval = FACT_INTERNAL_ParseWaveBank(
@@ -435,6 +527,14 @@ uint32_t FACTAudioEngine_CreateInMemoryWaveBank(
 		0,
 		ppWaveBank
 	);
+	if (pEngine->notifications & NOTIFY_WAVEBANKPREPARED)
+	{
+		note = (FACTNotification*) pEngine->pMalloc(sizeof(FACTNotification));
+		note->type = FACTNOTIFICATIONTYPE_WAVEBANKPREPARED;
+		note->waveBank.pWaveBank = *ppWaveBank;
+		note->pvContext = pEngine->wb_context;
+		LinkedList_AddEntry(&pEngine->wb_notifications_list, note, pEngine->apiLock, pEngine->pMalloc);
+	}
 	FAudio_PlatformUnlockMutex(pEngine->apiLock);
 	return retval;
 }
@@ -444,6 +544,7 @@ uint32_t FACTAudioEngine_CreateStreamingWaveBank(
 	const FACTStreamingParameters *pParms,
 	FACTWaveBank **ppWaveBank
 ) {
+	FACTNotification *note;
 	uint32_t retval, packetSize;
 	FAudio_PlatformLockMutex(pEngine->apiLock);
 	if (	pEngine->pReadFile == FACT_INTERNAL_DefaultReadFile &&
@@ -466,6 +567,14 @@ uint32_t FACTAudioEngine_CreateStreamingWaveBank(
 		1,
 		ppWaveBank
 	);
+	if (pEngine->notifications & NOTIFY_WAVEBANKPREPARED)
+	{
+		note = (FACTNotification*) pEngine->pMalloc(sizeof(FACTNotification));
+		note->type = FACTNOTIFICATIONTYPE_WAVEBANKPREPARED;
+		note->waveBank.pWaveBank = *ppWaveBank;
+		note->pvContext = pEngine->wb_context;
+		LinkedList_AddEntry(&pEngine->wb_notifications_list, note, pEngine->apiLock, pEngine->pMalloc);
+	}
 	FAudio_PlatformUnlockMutex(pEngine->apiLock);
 	return retval;
 }
@@ -1311,16 +1420,13 @@ uint32_t FACTSoundBank_Destroy(FACTSoundBank *pSoundBank)
 		FACTCue_Destroy(pSoundBank->cueList);
 	}
 
-	if (pSoundBank->parentEngine != NULL)
-	{
-		/* Remove this SoundBank from the Engine list */
-		LinkedList_RemoveEntry(
-			&pSoundBank->parentEngine->sbList,
-			pSoundBank,
-			pSoundBank->parentEngine->sbLock,
-			pSoundBank->parentEngine->pFree
-		);
-	}
+	/* Remove this SoundBank from the Engine list */
+	LinkedList_RemoveEntry(
+		&pSoundBank->parentEngine->sbList,
+		pSoundBank,
+		pSoundBank->parentEngine->sbLock,
+		pSoundBank->parentEngine->pFree
+	);
 
 	/* SoundBank Name */
 	pSoundBank->parentEngine->pFree(pSoundBank->name);
@@ -1487,16 +1593,13 @@ uint32_t FACTWaveBank_Destroy(FACTWaveBank *pWaveBank)
 		}
 	}
 
-	if (pWaveBank->parentEngine != NULL)
-	{
-		/* Remove this WaveBank from the Engine list */
-		LinkedList_RemoveEntry(
-			&pWaveBank->parentEngine->wbList,
-			pWaveBank,
-			pWaveBank->parentEngine->wbLock,
-			pWaveBank->parentEngine->pFree
-		);
-	}
+	/* Remove this WaveBank from the Engine list */
+	LinkedList_RemoveEntry(
+		&pWaveBank->parentEngine->wbList,
+		pWaveBank,
+		pWaveBank->parentEngine->wbLock,
+		pWaveBank->parentEngine->pFree
+	);
 
 	/* Free everything, finally. */
 	pWaveBank->parentEngine->pFree(pWaveBank->name);
@@ -1541,6 +1644,11 @@ uint32_t FACTWaveBank_Destroy(FACTWaveBank *pWaveBank)
 	}
 	FAudio_PlatformDestroyMutex(pWaveBank->waveLock);
 
+	if (pWaveBank->waveBankNames != NULL)
+	{
+		pWaveBank->parentEngine->pFree(pWaveBank->waveBankNames);
+	}
+
 	mutex = pWaveBank->parentEngine->apiLock;
 	pWaveBank->parentEngine->pFree(pWaveBank);
 	FAudio_PlatformUnlockMutex(mutex);
@@ -1596,8 +1704,26 @@ uint16_t FACTWaveBank_GetWaveIndex(
 	FACTWaveBank *pWaveBank,
 	const char *szFriendlyName
 ) {
-	FAudio_assert(0 && "WaveBank name tables are not supported!");
-	return 0;
+	uint16_t i;
+	char *curName;
+	if (pWaveBank == NULL || pWaveBank->waveBankNames == NULL)
+	{
+		return FACTINDEX_INVALID;
+	}
+
+	FAudio_PlatformLockMutex(pWaveBank->parentEngine->apiLock);
+	curName = pWaveBank->waveBankNames;
+	for (i = 0; i < pWaveBank->entryCount; i += 1, curName += 64)
+	{
+		if (FAudio_strncmp(szFriendlyName, curName, 64) == 0)
+		{
+			FAudio_PlatformUnlockMutex(pWaveBank->parentEngine->apiLock);
+			return i;
+		}
+	}
+	FAudio_PlatformUnlockMutex(pWaveBank->parentEngine->apiLock);
+
+	return FACTINDEX_INVALID;
 }
 
 uint32_t FACTWaveBank_GetWaveProperties(
@@ -1615,11 +1741,21 @@ uint32_t FACTWaveBank_GetWaveProperties(
 
 	entry = &pWaveBank->entries[nWaveIndex];
 
-	/* FIXME: Name tables! -flibit */
-	FAudio_zero(
-		pWaveProperties->friendlyName,
-		sizeof(pWaveProperties->friendlyName)
-	);
+	if (pWaveBank->waveBankNames)
+	{
+		FAudio_memcpy(
+			pWaveProperties->friendlyName,
+			&pWaveBank->waveBankNames[nWaveIndex * 64],
+			sizeof(pWaveProperties->friendlyName)
+		);
+	}
+	else
+	{
+		FAudio_zero(
+			pWaveProperties->friendlyName,
+			sizeof(pWaveProperties->friendlyName)
+		);
+	}
 
 	pWaveProperties->format = entry->Format;
 	pWaveProperties->durationInSamples = entry->PlayRegion.dwLength;
@@ -1636,6 +1772,10 @@ uint32_t FACTWaveBank_GetWaveProperties(
 			((entry->Format.wBlockAlign + 16) * 2)
 		);
 	}
+	else
+	{
+		FAudio_assert(0 && "Unrecognized wFormatTag!");
+	}
 
 	pWaveProperties->loopRegion = entry->LoopRegion;
 	pWaveProperties->streaming = pWaveBank->streaming;
@@ -1656,8 +1796,14 @@ uint32_t FACTWaveBank_Prepare(
 	FAudioBufferWMA bufferWMA;
 	FAudioVoiceSends sends;
 	FAudioSendDescriptor send;
-	FAudioADPCMWaveFormat format;
+	union
+	{
+		FAudioWaveFormatEx pcm;
+		FAudioADPCMWaveFormat adpcm;
+		FAudioXMA2WaveFormat xma2;
+	} format;
 	FACTWaveBankEntry *entry;
+	FACTSeekTable *seek;
 	if (pWaveBank == NULL)
 	{
 		*ppWave = NULL;
@@ -1702,38 +1848,78 @@ uint32_t FACTWaveBank_Prepare(
 	send.pOutputVoice = pWaveBank->parentEngine->master;
 	sends.SendCount = 1;
 	sends.pSends = &send;
-	format.wfx.nChannels = entry->Format.nChannels;
-	format.wfx.nSamplesPerSec = entry->Format.nSamplesPerSec;
+	format.pcm.nChannels = entry->Format.nChannels;
+	format.pcm.nSamplesPerSec = entry->Format.nSamplesPerSec;
 	if (entry->Format.wFormatTag == 0x0)
 	{
-		format.wfx.wFormatTag = FAUDIO_FORMAT_PCM;
-		format.wfx.wBitsPerSample = 8 << entry->Format.wBitsPerSample;
-		format.wfx.nBlockAlign = format.wfx.nChannels * format.wfx.wBitsPerSample / 8;
-		format.wfx.nAvgBytesPerSec = format.wfx.nBlockAlign * format.wfx.nSamplesPerSec;
-		format.wfx.cbSize = 0;
+		format.pcm.wFormatTag = FAUDIO_FORMAT_PCM;
+		format.pcm.wBitsPerSample = 8 << entry->Format.wBitsPerSample;
+		format.pcm.nBlockAlign = format.pcm.nChannels * format.pcm.wBitsPerSample / 8;
+		format.pcm.nAvgBytesPerSec = format.pcm.nBlockAlign * format.pcm.nSamplesPerSec;
+		format.pcm.cbSize = 0;
 	}
 	else if (entry->Format.wFormatTag == 0x1)
 	{
-		/* XMA2 is quite similar to WMA Pro. */
+		/* XMA2 is quite similar to WMA Pro... is what everyone thought.
+		 * What a great way to start this comment.
+		 *
+		 * Let's reconstruct the extra data because who knows what decoder we're dealing with in <present year>.
+		 * It's also a good exercise in understanding XMA2 metadata and feeding blocks into the decoder properly.
+		 * At the time of writing this patch, it's FFmpeg via gstreamer which doesn't even respect most of this.
+		 * ... which means: good luck to whoever ends up finding inaccuracies here in the future!
+		 *
+		 * dwLoopLength seems to match dwPlayLength in everything I've seen that had bLoopCount == 0.
+		 * dwLoopBegin can be > 0 even with bLoopCount == 0 because why not. Let's ignore that.
+		 *
+		 * dwSamplesEncoded is usually close to dwPlayLength but not always (if ever?) equal. Let's assume equality.
+		 * The XMA2 seek table uses sample indices as opposed to WMA's byte index seek table.
+		 *
+		 * nBlockAlign uses aWMABlockAlign given the entire WMA Pro thing BUT it's expected to be the block size for decoding.
+		 * The XMA2 block size MUST be a multiple of 2048 BUT entry->PlayRegion.dwLength / seek->entryCount doesn't respect that.
+		 * And even when correctly guesstimating the block size, we sometimes end up with block sizes >= 64k BYTES. nBlockAlign IS 16-BIT!
+		 * Scrap nBlockAlign. I've given up and made all FAudio gstreamer functions use dwBytesPerBlock if available.
+		 * Still though, if we don't want FAudio_INTERNAL_DecodeGSTREAMER to hang, the total data length must match (see SoundEffect.cs in FNA).
+		 * As such, we round up when guessing the block size, feed GStreamer with zeroes^Wundersized blocks and hope for the best.
+		 *
+		 * This is FUN.
+		 * -ade
+		 */
 		FAudio_assert(entry->Format.wBitsPerSample != 0);
 
-		format.wfx.wFormatTag = FAUDIO_FORMAT_XMAUDIO2;
-		format.wfx.nAvgBytesPerSec = aWMAAvgBytesPerSec[entry->Format.wBlockAlign >> 5];
-		format.wfx.nBlockAlign = aWMABlockAlign[entry->Format.wBlockAlign & 0x1F];
-		format.wfx.wBitsPerSample = 16;
-		format.wfx.cbSize = 0;
+		seek = &pWaveBank->seekTables[nWaveIndex];
+		format.pcm.wFormatTag = FAUDIO_FORMAT_XMAUDIO2;
+		format.pcm.wBitsPerSample = 16;
+		format.pcm.nAvgBytesPerSec = aWMAAvgBytesPerSec[entry->Format.wBlockAlign >> 5];
+		format.pcm.nBlockAlign = aWMABlockAlign[entry->Format.wBlockAlign & 0x1F];
+		format.pcm.cbSize = (
+			sizeof(FAudioXMA2WaveFormat) -
+			sizeof(FAudioWaveFormatEx)
+		);
+		format.xma2.wNumStreams = (format.pcm.nChannels + 1) / 2;
+		format.xma2.dwChannelMask = format.pcm.nChannels > 1 ? 0xFFFFFFFF >> (32 - format.pcm.nChannels) : 0;
+		format.xma2.dwSamplesEncoded = seek->entries[seek->entryCount - 1];
+		format.xma2.dwBytesPerBlock = (uint16_t) FAudio_ceil(
+			(double) entry->PlayRegion.dwLength /
+			(double) seek->entryCount /
+			2048.0
+		) * 2048;
+		format.xma2.dwPlayBegin = format.xma2.dwLoopBegin = 0;
+		format.xma2.dwPlayLength = format.xma2.dwLoopLength = format.xma2.dwSamplesEncoded;
+		format.xma2.bLoopCount = 0;
+		format.xma2.bEncoderVersion = 4;
+		format.xma2.wBlockCount = seek->entryCount;
 	}
 	else if (entry->Format.wFormatTag == 0x2)
 	{
-		format.wfx.wFormatTag = FAUDIO_FORMAT_MSADPCM;
-		format.wfx.nBlockAlign = (entry->Format.wBlockAlign + 22) * format.wfx.nChannels;
-		format.wfx.wBitsPerSample = 16;
-		format.wfx.cbSize = (
+		format.pcm.wFormatTag = FAUDIO_FORMAT_MSADPCM;
+		format.pcm.nBlockAlign = (entry->Format.wBlockAlign + 22) * format.pcm.nChannels;
+		format.pcm.wBitsPerSample = 16;
+		format.pcm.cbSize = (
 			sizeof(FAudioADPCMWaveFormat) -
 			sizeof(FAudioWaveFormatEx)
 		);
-		format.wSamplesPerBlock = (
-			((format.wfx.nBlockAlign / format.wfx.nChannels) - 6) * 2
+		format.adpcm.wSamplesPerBlock = (
+			((format.pcm.nBlockAlign / format.pcm.nChannels) - 6) * 2
 		);
 	}
 	else if (entry->Format.wFormatTag == 0x3)
@@ -1741,11 +1927,11 @@ uint32_t FACTWaveBank_Prepare(
 		/* Apparently this is used to detect WMA Pro...? */
 		FAudio_assert(entry->Format.wBitsPerSample == 0);
 
-		format.wfx.wFormatTag = FAUDIO_FORMAT_WMAUDIO2;
-		format.wfx.nAvgBytesPerSec = aWMAAvgBytesPerSec[entry->Format.wBlockAlign >> 5];
-		format.wfx.nBlockAlign = aWMABlockAlign[entry->Format.wBlockAlign & 0x1F];
-		format.wfx.wBitsPerSample = 16;
-		format.wfx.cbSize = 0;
+		format.pcm.wFormatTag = FAUDIO_FORMAT_WMAUDIO2;
+		format.pcm.nAvgBytesPerSec = aWMAAvgBytesPerSec[entry->Format.wBlockAlign >> 5];
+		format.pcm.nBlockAlign = aWMABlockAlign[entry->Format.wBlockAlign & 0x1F];
+		format.pcm.wBitsPerSample = 16;
+		format.pcm.cbSize = 0;
 	}
 	else
 	{
@@ -1761,11 +1947,11 @@ uint32_t FACTWaveBank_Prepare(
 	(*ppWave)->callback.callback.OnVoiceProcessingPassEnd = NULL;
 	(*ppWave)->callback.callback.OnVoiceProcessingPassStart = NULL;
 	(*ppWave)->callback.wave = *ppWave;
-	(*ppWave)->srcChannels = format.wfx.nChannels;
+	(*ppWave)->srcChannels = format.pcm.nChannels;
 	FAudio_CreateSourceVoice(
 		pWaveBank->parentEngine->audio,
 		&(*ppWave)->voice,
-		&format.wfx,
+		&format.pcm,
 		FAUDIO_VOICE_USEFILTER, /* FIXME: Can this be optional? */
 		4.0f,
 		(FAudioVoiceCallback*) &(*ppWave)->callback,
@@ -1775,19 +1961,19 @@ uint32_t FACTWaveBank_Prepare(
 	if (pWaveBank->streaming)
 	{
 		/* Init stream cache info */
-		if (format.wfx.wFormatTag == FAUDIO_FORMAT_PCM)
+		if (format.pcm.wFormatTag == FAUDIO_FORMAT_PCM)
 		{
 			(*ppWave)->streamSize = (
-				format.wfx.nSamplesPerSec *
-				format.wfx.nBlockAlign
+				format.pcm.nSamplesPerSec *
+				format.pcm.nBlockAlign
 			);
 		}
-		else if (format.wfx.wFormatTag == FAUDIO_FORMAT_MSADPCM)
+		else if (format.pcm.wFormatTag == FAUDIO_FORMAT_MSADPCM)
 		{
 			(*ppWave)->streamSize = (
-				format.wfx.nSamplesPerSec /
-				format.wSamplesPerBlock *
-				format.wfx.nBlockAlign
+				format.pcm.nSamplesPerSec /
+				format.adpcm.wSamplesPerBlock *
+				format.pcm.nBlockAlign
 			);
 		}
 		else
@@ -1797,7 +1983,7 @@ uint32_t FACTWaveBank_Prepare(
 
 			/* XACT does NOT support loop subregions for these formats */
 			FAudio_assert(entry->LoopRegion.dwStartSample == 0);
-			FAudio_assert(entry->LoopRegion.dwTotalSamples == entry->Duration);
+			FAudio_assert(entry->LoopRegion.dwTotalSamples == 0 || entry->LoopRegion.dwTotalSamples == entry->Duration);
 		}
 		(*ppWave)->streamCache = (uint8_t*) pWaveBank->parentEngine->pMalloc(
 			(*ppWave)->streamSize
@@ -1832,8 +2018,7 @@ uint32_t FACTWaveBank_Prepare(
 			buffer.LoopCount = nLoopCount;
 		}
 		buffer.pContext = NULL;
-		if (	format.wfx.wFormatTag == FAUDIO_FORMAT_WMAUDIO2 ||
-			format.wfx.wFormatTag == FAUDIO_FORMAT_XMAUDIO2	)
+		if (format.pcm.wFormatTag == FAUDIO_FORMAT_WMAUDIO2)
 		{
 			bufferWMA.pDecodedPacketCumulativeBytes =
 				pWaveBank->seekTables[nWaveIndex].entries;
@@ -2021,11 +2206,13 @@ uint32_t FACTWave_Stop(FACTWave *pWave, uint32_t dwFlags)
 	{
 		FACTNotification note;
 		note.type = FACTNOTIFICATIONTYPE_WAVESTOP;
+		note.wave.cueIndex = pWave->parentCue->index;
+		note.wave.pCue = pWave->parentCue;
+		note.wave.pSoundBank = pWave->parentCue->parentBank;
 		note.wave.pWave = pWave;
-		if (pWave->parentBank->parentEngine->notifications & NOTIFY_WAVESTOP)
-		{
-			note.pvContext = pWave->parentBank->parentEngine->wave_context;
-		}
+		note.wave.pWaveBank = pWave->parentBank;
+		note.pvContext = pWave->parentBank->parentEngine->wave_context;
+
 		pWave->parentBank->parentEngine->notificationCallback(&note);
 	}
 
@@ -2227,7 +2414,6 @@ uint32_t FACTCue_Destroy(FACTCue *pCue)
 {
 	FACTCue *cue, *prev;
 	FAudioMutex mutex;
-	FACTNotification note;
 	if (pCue == NULL)
 	{
 		return 1;
@@ -2261,20 +2447,7 @@ uint32_t FACTCue_Destroy(FACTCue *pCue)
 	FAudio_assert(cue != NULL && "Could not find Cue reference!");
 
 	pCue->parentBank->parentEngine->pFree(pCue->variableValues);
-	if (pCue->notifyOnDestroy || pCue->parentBank->parentEngine->notifications & NOTIFY_CUEDESTROY)
-	{
-		note.type = FACTNOTIFICATIONTYPE_CUEDESTROYED;
-		note.cue.pCue = pCue;
-		if (pCue->parentBank->parentEngine->notifications & NOTIFY_CUEDESTROY)
-		{
-			note.pvContext = pCue->parentBank->parentEngine->cue_context;
-		}
-		else
-		{
-			note.pvContext = pCue->usercontext;
-		}
-		pCue->parentBank->parentEngine->notificationCallback(&note);
-	}
+	FACT_INTERNAL_SendCueNotification(pCue, NOTIFY_CUEDESTROY, FACTNOTIFICATIONTYPE_CUEDESTROYED);
 
 	mutex = pCue->parentBank->parentEngine->apiLock;
 	pCue->parentBank->parentEngine->pFree(pCue);
@@ -2316,6 +2489,9 @@ uint32_t FACTCue_Play(FACTCue *pCue)
 				FACT_STATE_STOPPING |
 				FACT_STATE_PAUSED
 			);
+
+			FACT_INTERNAL_SendCueNotification(pCue, NOTIFY_CUESTOP, FACTNOTIFICATIONTYPE_CUESTOP);
+
 			FAudio_PlatformUnlockMutex(
 				pCue->parentBank->parentEngine->apiLock
 			);
@@ -2414,6 +2590,9 @@ uint32_t FACTCue_Play(FACTCue *pCue)
 		FACT_STATE_STOPPED |
 		FACT_STATE_PREPARED
 	);
+
+	FACT_INTERNAL_SendCueNotification(pCue, NOTIFY_CUEPLAY, FACTNOTIFICATIONTYPE_CUEPLAY);
+
 	pCue->start = FAudio_timems();
 
 	/* If it's a simple wave, just play it! */
@@ -2512,10 +2691,15 @@ uint32_t FACTCue_Stop(FACTCue *pCue, uint32_t dwFlags)
 				pCue->maxRpcReleaseTime
 			);
 		}
-
-		pCue->state |= FACT_STATE_STOPPING;
+		else
+		{
+			/* Pretty sure this doesn't happen, but just in case? */
+			pCue->state |= FACT_STATE_STOPPING;
+		}
 	}
 
+	FACT_INTERNAL_SendCueNotification(pCue, NOTIFY_CUESTOP, FACTNOTIFICATIONTYPE_CUESTOP);
+
 	FAudio_PlatformUnlockMutex(pCue->parentBank->parentEngine->apiLock);
 	return 0;
 }
@@ -2583,6 +2767,8 @@ uint32_t FACTCue_SetMatrixCoefficients(
 		}
 	}
 
+	FACT_INTERNAL_SendCueNotification(pCue, NOTIFY_CUESTOP, FACTNOTIFICATIONTYPE_CUESTOP);
+
 	FAudio_PlatformUnlockMutex(pCue->parentBank->parentEngine->apiLock);
 	return 0;
 }
