@@ -1,6 +1,6 @@
 /* FAudio - XAudio Reimplementation for FNA
  *
- * Copyright (c) 2011-2021 Ethan Lee, Luigi Auriemma, and the MonoGame Team
+ * Copyright (c) 2011-2023 Ethan Lee, Luigi Auriemma, and the MonoGame Team
  *
  * This software is provided 'as-is', without any express or implied warranty.
  * In no event will the authors be held liable for any damages arising from
@@ -678,16 +678,27 @@ uint32_t FAudio_CreateMasteringVoice(
 	uint32_t DeviceIndex,
 	const FAudioEffectChain *pEffectChain
 ) {
-	FAudioDeviceDetails details;
-
 	LOG_API_ENTER(audio)
 
 	/* For now we only support one allocated master voice at a time */
 	FAudio_assert(audio->master == NULL);
 
-	if (FAudio_GetDeviceDetails(audio, DeviceIndex, &details) != 0)
+	if (	InputChannels == FAUDIO_DEFAULT_CHANNELS ||
+		InputSampleRate == FAUDIO_DEFAULT_SAMPLERATE	)
 	{
-		return FAUDIO_E_INVALID_CALL;
+		FAudioDeviceDetails details;
+		if (FAudio_GetDeviceDetails(audio, DeviceIndex, &details) != 0)
+		{
+			return FAUDIO_E_INVALID_CALL;
+		}
+		if (InputChannels == FAUDIO_DEFAULT_CHANNELS)
+		{
+			InputChannels = details.OutputFormat.Format.nChannels;
+		}
+		if (InputSampleRate == FAUDIO_DEFAULT_SAMPLERATE)
+		{
+			InputSampleRate = details.OutputFormat.Format.nSamplesPerSec;
+		}
 	}
 
 	*ppMasteringVoice = (FAudioMasteringVoice*) audio->pMalloc(sizeof(FAudioVoice));
@@ -704,12 +715,8 @@ uint32_t FAudio_CreateMasteringVoice(
 	(*ppMasteringVoice)->volume = 1.0f;
 
 	/* Master Properties */
-	(*ppMasteringVoice)->master.inputChannels = (InputChannels == FAUDIO_DEFAULT_CHANNELS) ?
-		details.OutputFormat.Format.nChannels :
-		InputChannels;
-	(*ppMasteringVoice)->master.inputSampleRate = (InputSampleRate == FAUDIO_DEFAULT_SAMPLERATE) ?
-		details.OutputFormat.Format.nSamplesPerSec :
-		InputSampleRate;
+	(*ppMasteringVoice)->master.inputChannels = InputChannels;
+	(*ppMasteringVoice)->master.inputSampleRate = InputSampleRate;
 
 	/* Sends/Effects */
 	FAudio_zero(&(*ppMasteringVoice)->sends, sizeof(FAudioVoiceSends));
