@@ -1,6 +1,6 @@
 /* FAudio - XAudio Reimplementation for FNA
  *
- * Copyright (c) 2011-2021 Ethan Lee, Luigi Auriemma, and the MonoGame Team
+ * Copyright (c) 2011-2023 Ethan Lee, Luigi Auriemma, and the MonoGame Team
  *
  * This software is provided 'as-is', without any express or implied warranty.
  * In no event will the authors be held liable for any damages arising from
@@ -24,12 +24,14 @@
  *
  */
 
+#ifndef FAUDIO_WIN32_PLATFORM
+
 #include "FAudio_internal.h"
 
 #include <SDL.h>
 
-#if !SDL_VERSION_ATLEAST(2, 0, 9)
-#error "SDL version older than 2.0.9"
+#if !SDL_VERSION_ATLEAST(2, 24, 0)
+#error "SDL version older than 2.24.0"
 #endif /* !SDL_VERSION_ATLEAST */
 
 /* Mixer Thread */
@@ -50,8 +52,47 @@ static void FAudio_INTERNAL_MixCallback(void *userdata, Uint8 *stream, int len)
 
 /* Platform Functions */
 
+static void FAudio_INTERNAL_PrioritizeDirectSound()
+{
+	int numdrivers, i, wasapi, directsound;
+
+	if (SDL_GetHint("SDL_AUDIODRIVER") != NULL)
+	{
+		/* Already forced to something, ignore */
+		return;
+	}
+
+	/* Check to see if we have both Windows drivers in the list */
+	numdrivers = SDL_GetNumAudioDrivers();
+	wasapi = -1;
+	directsound = -1;
+	for (i = 0; i < numdrivers; i += 1)
+	{
+		const char *driver = SDL_GetAudioDriver(i);
+		if (SDL_strcmp(driver, "wasapi") == 0)
+		{
+			wasapi = i;
+		}
+		else if (SDL_strcmp(driver, "directsound") == 0)
+		{
+			directsound = i;
+		}
+	}
+
+	/* We force if and only if both drivers exist and wasapi is earlier */
+	if ((wasapi > -1) && (directsound > -1))
+	{
+		if (wasapi < directsound)
+		{
+			SDL_SetHint("SDL_AUDIODRIVER", "directsound");
+		}
+	}
+}
+
 void FAudio_PlatformAddRef()
 {
+	FAudio_INTERNAL_PrioritizeDirectSound();
+
 	/* SDL tracks ref counts for each subsystem */
 	if (SDL_InitSubSystem(SDL_INIT_AUDIO) < 0)
 	{
@@ -79,8 +120,6 @@ void FAudio_PlatformInit(
 ) {
 	SDL_AudioDeviceID device;
 	SDL_AudioSpec want, have;
-	const char *driver;
-	int changes = 0;
 
 	FAudio_assert(mixFormat != NULL);
 	FAudio_assert(updateSize != NULL);
@@ -106,48 +145,6 @@ void FAudio_PlatformInit(
 		want.samples = want.freq / 100;
 	}
 
-	/* FIXME: SDL bug!
-	 * The PulseAudio backend does this annoying thing where it halves the
-	 * buffer size to prevent latency issues:
-	 *
-	 * https://hg.libsdl.org/SDL/file/df343364c6c5/src/audio/pulseaudio/SDL_pulseaudio.c#l577
-	 *
-	 * To get the _actual_ quantum size we want, we just double the buffer
-	 * size and allow SDL to set the quantum size back to normal.
-	 * -flibit
-	 */
-	driver = SDL_GetCurrentAudioDriver();
-	if (SDL_strcmp(driver, "pulseaudio") == 0)
-	{
-		want.samples *= 2;
-		changes = SDL_AUDIO_ALLOW_SAMPLES_CHANGE;
-	}
-
-	/* FIXME: SDL bug!
-	 * The most common backends support varying samples values, but many
-	 * require a power-of-two value, which XAudio2 is not a fan of.
-	 * Normally SDL creates an intermediary stream to handle this, but this
-	 * has not been written yet:
-	 * https://bugzilla.libsdl.org/show_bug.cgi?id=5136
-	 * -flibit
-	 */
-	else if (	SDL_strcmp(driver, "emscripten") == 0 ||
-			SDL_strcmp(driver, "dsp") == 0	)
-	{
-		want.samples -= 1;
-		want.samples |= want.samples >> 1;
-		want.samples |= want.samples >> 2;
-		want.samples |= want.samples >> 4;
-		want.samples |= want.samples >> 8;
-		want.samples |= want.samples >> 16;
-		want.samples += 1;
-		SDL_Log(
-			"Forcing FAudio quantum to a power-of-two.\n"
-			"You don't actually want this, it's technically a bug:\n"
-			"https://bugzilla.libsdl.org/show_bug.cgi?id=5136"
-		);
-	}
-
 	/* Open the device (or at least try to) */
 iosretry:
 	device = SDL_OpenAudioDevice(
@@ -155,7 +152,7 @@ iosretry:
 		0,
 		&want,
 		&have,
-		changes
+		0
 	);
 	if (device == 0)
 	{
@@ -221,9 +218,13 @@ uint32_t FAudio_PlatformGetDeviceDetails(
 ) {
 	const char *name, *envvar;
 	int channels, rate;
+	SDL_AudioSpec spec;
+	uint32_t devcount;
 
 	FAudio_zero(details, sizeof(FAudioDeviceDetails));
-	if (index >= FAudio_PlatformGetDeviceCount())
+
+	devcount = FAudio_PlatformGetDeviceCount();
+	if (index >= devcount)
 	{
 		return FAUDIO_E_INVALID_CALL;
 	}
@@ -258,17 +259,59 @@ uint32_t FAudio_PlatformGetDeviceDetails(
 		sizeof(details->DisplayName)
 	);
 
-	/* TODO: SDL_GetAudioDeviceSpec! */
+	/* Environment variables take precedence over all possible values */
 	envvar = SDL_getenv("SDL_AUDIO_FREQUENCY");
-	if (!envvar || ((rate = SDL_atoi(envvar)) == 0))
+	if (envvar != NULL)
+	{
+		rate = SDL_atoi(envvar);
+	}
+	else
+	{
+		rate = 0;
+	}
+	envvar = SDL_getenv("SDL_AUDIO_CHANNELS");
+	if (envvar != NULL)
+	{
+		channels = SDL_atoi(envvar);
+	}
+	else
+	{
+		channels = 0;
+	}
+
+	/* Get the device format from the OS */
+	if (index == 0)
+	{
+		/* TODO: Do we want to squeeze the name into the output? */
+		if (SDL_GetDefaultAudioInfo(NULL, &spec, 0) < 0)
+		{
+			SDL_zero(spec);
+		}
+	}
+	else
+	{
+		SDL_GetAudioDeviceSpec(index - 1, 0, &spec);
+	}
+	if ((spec.freq > 0) && (rate <= 0))
+	{
+		rate = spec.freq;
+	}
+	if ((spec.channels > 0) && (channels <= 0))
+	{
+		channels = spec.channels;
+	}
+
+	/* If we make it all the way here with no format, hardcode a sane one */
+	if (rate <= 0)
 	{
 		rate = 48000;
 	}
-	envvar = SDL_getenv("SDL_AUDIO_CHANNELS");
-	if (!envvar || ((channels = SDL_atoi(envvar)) == 0))
+	if (channels <= 0)
 	{
 		channels = 2;
 	}
+
+	/* Write the format, finally. */
 	WriteWaveFormatExtensible(
 		&details->OutputFormat,
 		channels,
@@ -600,3 +643,9 @@ void FAudio_UTF8_To_UTF16(const char *src, uint16_t *dst, size_t len)
 }
 
 /* vim: set noexpandtab shiftwidth=8 tabstop=8: */
+
+#else
+
+extern int this_tu_is_empty;
+
+#endif /* FAUDIO_WIN32_PLATFORM */
