@@ -1,6 +1,6 @@
 /* FAudio - XAudio Reimplementation for FNA
  *
- * Copyright (c) 2011-2022 Ethan Lee, Luigi Auriemma, and the MonoGame Team
+ * Copyright (c) 2011-2023 Ethan Lee, Luigi Auriemma, and the MonoGame Team
  *
  * This software is provided 'as-is', without any express or implied warranty.
  * In no event will the authors be held liable for any damages arising from
@@ -55,11 +55,6 @@ std::vector<bool> wavebankShows;
 
 std::vector<FACTWave*> waves;
 
-void FAudioTool_Init()
-{
-	/* Nothing to do... */
-}
-
 void FAudioTool_Quit()
 {
 	for (size_t i = 0; i < soundBanks.size(); i += 1)
